
from validators.abstract_validator import AbsValidator


class Fastq(AbsValidator):
    """The class to validate the Fastq format."""

    def validate_file(self, file_path):
        """Validates whether the file at the given path is of the fastq format or not.
            (also see AbsValidator.)

            :returns boolean whether the format is valid
                string message why it is not valid or an empty string.
        """

        with open(file_path, "r") as file:

            count = 0
            len_of_sequence = -1

            for line in file:
                line = line.strip()

                if line == "":
                    return False, "Empty line at line: " + str(count+1)

                line_kind = count % 4
                valid_line, msg = self.test_line(line, line_kind)

                if not valid_line:
                    return False, msg + " Line: " + str(count+1)

                if line_kind == 1:
                    len_of_sequence = len(line)
                elif line_kind == 3:
                    if len(line) != len_of_sequence:
                        return False, "The quality sequence is not the same length as the DNA sequence. Line:" + str(count+1)

                count += 1

            if count % 4 != 0:
                return False, "Last entry is missing information. Line:" + str(count+1)

        return True, ""

    def test_line(self, line, kind):
        """calls the corresponding test depending on the given kind.

            kind:
            0 - header
            1 - DNA sequence
            2 - Plus-line between sequence and quality
            3 - Quality sequence

        """
        if kind == 0:
            return line.startswith("@"), "Expected header line does not start with '@'."
        elif kind == 1:
            return self.test_sequence(line), "Expected DNA sequence contains banned characters."
        elif kind == 2:
            return line.startswith('+'), "Expected Plus-line does not start with a '+'."
        elif kind == 3:
            return self.test_quality(line), "Expected quality sequence contains banned characters."

    @staticmethod
    def test_sequence(line):
        """tests the given line to only contain characters in the set [A, C, T, G]"""
        for c in line:
            if c not in ['A', 'C', 'T', 'G']:
                return False
        return True

    def test_quality(self, line):
        """tests the given line's characters to be in the set of allowed symbols defined in get_quality_symbols()."""
        symbols = self.get_quality_symbols()
        for c in line:
            if c not in symbols:
                return False
        return True

    @staticmethod
    def get_quality_symbols():
        """returns a list of all allowed quality symbols in asc order:
            lowest--->highest
            !"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~]
        """

        all_string = "!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~]"
        return list(all_string)




class AbsValidator:
    """The parent class for all validators.
        It defines the functions that all validator plugins have to implement and overwrite.
    """

    def validate_file(self, file_path):
        """ The function validate_file is the main workhorse and connection to the rest of the project.
            As its name indicates it is the function used to validate the files.

            It has two return values:
                the first is a boolean of whether the file is valid
                the second is a string. If the file is valid it should be empty
                                        If the file is corrupt it should be an explanation why(if possible)

         """
        valid = False
        return valid, "Error: Unimplemented validate_file function."




from validators.abstract_validator import AbsValidator


class Fasta(AbsValidator):
    """Fasta format validator following the NCBI definition.

    Validates whether the file at the given path is of the fasta format or not.
    Depending on the given code at the initialization it tests for DNA or aminoacid coding.

    The format definition can be found here:
    https://blast.ncbi.nlm.nih.gov/Blast.cgi?CMD=Web&PAGE_TYPE=BlastDocs&DOC_TYPE=BlastHelp (17.6.18)
    """

    def __init__(self, code):
        """Constructor for the fasta-Validator the given code decides whether it tests for DNA or aminoacid coding."""

        if code == "DNA":
            self.allowed = self.get_dna()
        if code == "AA":
            self.allowed = self.get_aa()

    def validate_file(self, file_path):
        """Validates whether the file at the given path is of the fastq format or not.
            (also see AbsValidator.)

            :returns boolean whether the format is valid,
                    string message why it is not or an empty string if valid.
        """

        # sorry for the indent heaven.
        with open(file_path, "r") as file:

            had_header = False
            count = 1
            for line in file:
                line = line.strip()
                if line == "":
                    return False, "Empty line at line: " + str(count)

                if had_header:
                    if line.startswith(">"):
                        return False, "Header without a sequence in line: " + str(count-1)

                    result, message, char_pos = self.validate_line_coding(line)

                    if result:
                        had_header = False
                        count += 1
                    else:
                        return False, message+" At line: " + str(count) + ":" + str(char_pos)

                else:
                    if line.startswith(">"):
                        had_header = True
                        count += 1
                        continue
                    else:
                        result, message, char_pos = self.validate_line_coding(line)

                        if result:
                            count += 1
                        else:
                            return False, message + " at line: " + str(count) + ":" + str(char_pos)

            if had_header:
                return False, "Header at end of file with no sequence."

            return True, ""

    def validate_line_coding(self, line):
        """checks the given line for characters not allowed by the format either DNA or aminoacid coding
            returns first whether the line coding was correct
                second a message containing the char that was wrong if there was one else an empty string
                third the position at which the faulty char was found
        """
        c_count = 1
        for c in line:
            if c in self.allowed:
                c_count += 1
                continue
            else:
                return False, "Character not allowed [" + c + "] in sequence.", c_count

        return True, "", -1

    """
        A  adenosine          C  cytidine             G  guanine
        T  thymidine          N  A/G/C/T (any)        U  uridine 
        K  G/T (keto)         S  G/C (strong)         Y  T/C (pyrimidine) 
        M  A/C (amino)        W  A/T (weak)           R  G/A (purine)        
        B  G/T/C              D  G/A/T                H  A/C/T      
        V  G/C/A              -  gap of indeterminate length
    """

    @staticmethod
    def get_dna():
        """returns a list of all allowed characters coding for a dna base"""
        return ['A', 'C', 'G', 'T', 'N', 'U', 'K', 'S', 'Y', 'M', 'W', 'R', 'B', 'D', 'H', 'V', '-']

    """
        A  alanine               P  proline       
        B  aspartate/asparagine  Q  glutamine      
        C  cystine               R  arginine      
        D  aspartate             S  serine      
        E  glutamate             T  threonine      
        F  phenylalanine         U  selenocysteine
        G  glycine               V  valine        
        H  histidine             W  tryptophan        
        I  isoleucine            Y  tyrosine
        K  lysine                Z  glutamate/glutamine
        L  leucine               X  any
        M  methionine            *  translation stop
        N  asparagine            -  gap of indeterminate length
    """

    @staticmethod
    def get_aa():
        """returns a list of all allowed characters coding for an aminoacid"""
        return ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L', 'M', 'N', 'P', 'Q', 'R', 'S', 'T', 'U', 'V',
                'W', 'Y', 'Z', 'X', '*', '-']




from validators.fasta_validator import Fasta
from validators.fastq_validator import Fastq


def available_formats():
    """Returns a list of all supported file-types"""
    return ["fasta-dna", "fasta-aa", "fastq"]


def get_format_from_ending(file_ending):
    """Returns the corresponding format-name to the given file-ending and an empty string if the ending is unknown."""
    if file_ending == "fasta":
        return "fasta-dna"
    if file_ending == "fastq":
        return "fastq"

    # in case of an unknown ending return empty string.
    return ""


def get_validator(name):
    """Returns the corresponding validator to the given format-name"""
    if name == "fasta-dna":
        return Fasta("DNA")
    if name == "fasta-aa":
        return Fasta("AA")
    if name == "fastq":
        return Fastq()

    raise Exception("FormatFactory.get_format() - Error: Unknown Format identifier.")


def get_uncertain_endings():
    """returns a list of all file endings that don't fully define their format.
        For formats that for example can contain different kinds of data
        like .fasta with genetic and aminoacid sequences.
    """
    return ["fasta"]



#!/usr/bin/env python

from setuptools import setup, find_packages

version = '1.0.0'

with open('README.md') as f:
    readme = f.read()

with open('LICENSE') as f:
    license = f.read()

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name='Valifor',
    version=version,
    description='A small validator for formats in the NCBI context',
    long_description=readme,
    keywords=['validator', 'formats', 'bioinformatics', 'fasta', 'fastq'],
    author='Najia Ahmadi and Alexander Ott',
    author_email="a.ott@student.uni-tuebingen.de",
    url="https://github.com/qbicsoftware/dmqb-grp1-2018",
    license=license,
    scripts=['scripts/valifor'],  # entry point! name of the script also sets name of the command in the cmd-line.
    install_requires=required,
    packages=find_packages(exclude='docs'),
    include_package_data=True
)
